#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>

#define MAX_STR 255
#define MAX_CASES 1000
#define MAXTRF 30       /* numero maximo de tarefas */

void* tarefa(void* arg);        

int casos_teste[MAX_CASES][2];   // argumentos de cada caso de teste
int resultados[MAX_CASES][2]; // resultados esperados e observados
int totalCasos, sucessos, falhas, casosProcessados; // variveis de controlo e finais
pthread_mutex_t mtx_casos, mtx_sucessos, mtx_falhas; /* mutexes */


int Combinacoes(int n, int r)
{
    int i, combinacoes;

    if (n < r || r < 1)
        return 0;

    i = 1;
    combinacoes = 1;
    while (i <= r)
    {
        combinacoes *= (n - r + i);
        combinacoes /= i;
        i++;
    }
    return combinacoes;
}

int LerCasos(char *casos) {
    FILE* f;
    char str[MAX_STR], *pt;
    casosProcessados = 0;
    totalCasos = 0;
    sucessos = 0;
    falhas = 0;
    if ((f = fopen(casos, "rt")) != NULL) {
        while (fgets(str, MAX_STR, f) != NULL) {
            pt = strtok(str, " ");
            casos_teste[totalCasos][0] = atoi(pt);
            pt = strtok(NULL, " ");
            casos_teste[totalCasos][1] = atoi(pt);
            pt = strtok(NULL, " ");
            resultados[totalCasos][0] = atoi(pt);
            totalCasos++;
            if (totalCasos >= MAX_CASES - 1)
                break;
        }
        fclose(f);
        return 1;
    }
    printf("\nErro ao ler o ficheiro de casos.");
    return 0;
}


int main(int argc, char **argv)
{
    int i, r, N;
    pthread_t      trfid[MAXTRF];    /* variavel para ID das tarefas */
    pthread_attr_t trfatr;           /* variavel para atributos das tarefas */

    N = atoi(argv[1]);
    if (N < 1)
        N = 1;

    // ler dados
    if (!LerCasos(argv[2])) // atualiza asvariveis globais
        exit(1);

    /* inicializar variavel de atributos com valores por defeito */
    pthread_attr_init(&trfatr);
    /* modificar estado de desacoplamento para "joinable" */
    pthread_attr_setdetachstate(&trfatr, PTHREAD_CREATE_JOINABLE);
    /* inicializar mutex com valores por defeito para os atributos */
    pthread_mutex_init(&mtx_casos, NULL);
    pthread_mutex_init(&mtx_sucessos, NULL);
    pthread_mutex_init(&mtx_falhas, NULL);

    /* inicializar estruturas de dados que servem de argumento das tarefas */
    for (i = 0; i < N; i++) {
        /* criar e iniciar execucao de tarefa */
        r = pthread_create(&trfid[i], &trfatr, tarefa, NULL);
        if (r) {
            /* erro ! */
            perror("Erro na criacao da tarefa!");
            exit(1);
        }
    }

    /* esperar que as tarefas criadas terminem */
    for (i = 0; i < N; i++)
        pthread_join(trfid[i], (void**)NULL);
    /* libertar recursos associados ao mutex */
    pthread_mutex_destroy(&mtx_casos);
    pthread_mutex_destroy(&mtx_sucessos);
    pthread_mutex_destroy(&mtx_falhas);

    printf("\nResultados: \nSucesso: %d\nFalhas: %d\nTotal casos: %d.", 
        sucessos, falhas, totalCasos);

    for (i = 0; i < totalCasos; i++)
        if (resultados[i][0] != resultados[i][1])
            printf("\nErro no caso %d: input (%d, %d) esperado/observado: %d / %d.",
                i, casos_teste[i][0], casos_teste[i][1],
                resultados[i][0], resultados[i][1]);

    return 0;
}

void* tarefa(void* arg)
{
    int item=-1;
    do {
        // obter um valor para processar
        pthread_mutex_lock(&mtx_casos);
        item = casosProcessados++;
        pthread_mutex_unlock(&mtx_casos);
        if (item < totalCasos) {
            // executa e grava o resultado observado
            if ((resultados[item][1]=Combinacoes(casos_teste[item][0], casos_teste[item][1]))
                == resultados[item][0]) 
            {
                // sucesso
                pthread_mutex_lock(&mtx_sucessos);
                sucessos++;
                pthread_mutex_unlock(&mtx_sucessos);
            }
            else {
                // insucesso
                pthread_mutex_lock(&mtx_falhas);
                falhas++;
                pthread_mutex_unlock(&mtx_falhas);
            }
        } 
    } while (item<totalCasos);
    return (void*)NULL;
}
