#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>

#define MAX_STR 255
#define MAX_CASES 1000

void* tarefa(void* arg);        

typedef struct {
    int *casos_teste[2];   // argumentos de cada caso de teste
    int *resultados[2]; // resultados esperados e observados
    int totalCasos, sucessos, falhas, casosProcessados; // variveis de controlo e finais
    pthread_mutex_t mtx_casos, mtx_sucessos, mtx_falhas; /* mutexes */
} TMemTarefas;

// chamar aps totalCasos estar preenchido
void AlocarMemTarefas(TMemTarefas* mem)
{
    int i;
    for (i = 0; i < 2; i++) {
        mem->casos_teste[i] = (int*)malloc(mem->totalCasos * sizeof(int));
        mem->resultados[i] = (int*)malloc(mem->totalCasos * sizeof(int));
        if (mem->casos_teste[i] == NULL ||
            mem->resultados[i] == NULL)
            exit(1);
    }
}

void LibertarMemTarefas(TMemTarefas* mem)
{
    int i;
    for (i = 0; i < 2; i++) {
        free(mem->casos_teste[i]);
        free(mem->resultados[i]);
    }
}


int Combinacoes(int n, int r)
{
    int i, combinacoes;

    if (n < r || r < 1)
        return 0;

    i = 1;
    combinacoes = 1;
    while (i <= r)
    {
        combinacoes *= (n - r + i);
        combinacoes /= i;
        i++;
    }
    return combinacoes;
}

int LerCasos(TMemTarefas*mem, char *casos) {
    FILE* f;
    char str[MAX_STR], *pt;
    int i;
    mem->casosProcessados = 0;
    mem->totalCasos = 0;
    mem->sucessos = 0;
    mem->falhas = 0;
    if ((f = fopen(casos, "rt")) != NULL) {
        // duas passagens pelo ficheiro, primeira passagem para contar as linhas
        while (fgets(str, MAX_STR, f) != NULL) 
            mem->totalCasos++;
        // alocar memria
        AlocarMemTarefas(mem);

        fseek(f, 0, SEEK_SET); // retomar para o incio
        i = 0;
        while (fgets(str, MAX_STR, f) != NULL) {
            pt = strtok(str, " ");
            mem->casos_teste[0][i] = atoi(pt);
            pt = strtok(NULL, " ");
            mem->casos_teste[1][i] = atoi(pt);
            pt = strtok(NULL, " ");
            mem->resultados[0][i] = atoi(pt);
            i++;
        }
        fclose(f);
        return 1;
    }
    printf("\nErro ao ler o ficheiro de casos.");
    return 0;
}



int main(int argc, char **argv)
{
    int i, r, N;
    TMemTarefas mem;
    pthread_t      *trfid;           /* variavel para ID das tarefas */
    pthread_attr_t trfatr;           /* variavel para atributos das tarefas */

    N = atoi(argv[1]);
    if (N < 1)
        N = 1;

    // ler dados
    if (!LerCasos(&mem, argv[2])) // atualiza asvariveis globais
        exit(1);

    trfid = (pthread_t*)malloc(N * sizeof(pthread_t));
    if (trfid == NULL)
        exit(1);

    /* inicializar variavel de atributos com valores por defeito */
    pthread_attr_init(&trfatr);
    /* modificar estado de desacoplamento para "joinable" */
    pthread_attr_setdetachstate(&trfatr, PTHREAD_CREATE_JOINABLE);
    /* inicializar mutex com valores por defeito para os atributos */
    pthread_mutex_init(&mem.mtx_casos, NULL);
    pthread_mutex_init(&mem.mtx_sucessos, NULL);
    pthread_mutex_init(&mem.mtx_falhas, NULL);

    /* inicializar estruturas de dados que servem de argumento das tarefas */
    for (i = 0; i < N; i++) {
        /* criar e iniciar execucao de tarefa */
        r = pthread_create(&trfid[i], &trfatr, tarefa, (void*) & mem);
        if (r) {
            /* erro ! */
            perror("Erro na criacao da tarefa!");
            exit(1);
        }
    }

    /* esperar que as tarefas criadas terminem */
    for (i = 0; i < N; i++)
        pthread_join(trfid[i], (void**)NULL);
    /* libertar recursos associados ao mutex */
    pthread_mutex_destroy(&mem.mtx_casos);
    pthread_mutex_destroy(&mem.mtx_sucessos);
    pthread_mutex_destroy(&mem.mtx_falhas);

    printf("\nResultados: \nSucesso: %d\nFalhas: %d\nTotal casos: %d.", 
        mem.sucessos, mem.falhas, mem.totalCasos);

    for (i = 0; i < mem.totalCasos; i++)
        if (mem.resultados[0][i] != mem.resultados[1][i])
            printf("\nErro no caso %d: input (%d, %d) esperado/observado: %d / %d.",
                i, mem.casos_teste[0][i], mem.casos_teste[1][i],
                mem.resultados[0][i], mem.resultados[1][i]);

    free(trfid);
    LibertarMemTarefas(&mem);
    return 0;
}

void* tarefa(void* arg)
{
    int item = -1;
    TMemTarefas* mem = (TMemTarefas*)arg;
    do {
        // obter um valor para processar
        pthread_mutex_lock(&mem->mtx_casos);
        item = mem->casosProcessados++;
        pthread_mutex_unlock(&mem->mtx_casos);
        if (item < mem->totalCasos) {
            // executa e grava o resultado observado
            if ((mem->resultados[1][item] = Combinacoes(mem->casos_teste[0][item], mem->casos_teste[1][item]))
                == mem->resultados[0][item])
            {
                // sucesso
                pthread_mutex_lock(&mem->mtx_sucessos);
                mem->sucessos++;
                pthread_mutex_unlock(&mem->mtx_sucessos);
            }
            else {
                // insucesso
                pthread_mutex_lock(&mem->mtx_falhas);
                mem->falhas++;
                pthread_mutex_unlock(&mem->mtx_falhas);
            }
        }
    } while (item < mem->totalCasos);
    return (void*)NULL;
}
